/*
  inverse function of log(a) - psi(a)
*/

#ifndef INV_LOGA_PSIA_H
#define INV_LOGA_PSIA_H

#ifdef __cplusplus
extern "C" {
#endif

  double inv_log_psi(double v);

#ifdef __cplusplus
}
#endif

#endif
